/* 
  Copyright (c) 2003. The Regents of the University of California (Regents).
                      All Rights Reserved.
  
  Created by Gavin E. Crooks <gec@compbio.berkeley.edu>, University of
  California, Berkeley.
  
  Permission is hereby granted, free of charge, to any person obtaining a 
  copy of this software and associated documentation files (the "Software"),
  to deal in the Software without restriction, including without limitation
  the rights to use, copy, modify, merge, publish, distribute, sublicense,
  and/or sell copies of the Software, and to permit persons to whom the
  Software is furnished to do so, subject to the following conditions:

  The above copyright notice and this permission notice shall be included
  in all copies or substantial portions of the Software.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING 
  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER 
  DEALINGS IN  THE SOFTWARE.

  (This is the BSD/MIT Open Source License, 
  http://www.opensource.org/licenses/mit-license.html)
*/


/* **********************************************************
  second-hmm : 
   Predict protein secondary structure from an amino acid 
   sequence using a hidden Markov model.

   Gavin E. Crooks <gec@compbio.berkeley.edu>

   try "second-hmm --help" for usage information.

 *********************************************************** */

/*  ---THINGS TODO ---
  Document datafiles.
  Fix handling of hmm files
*/

#include <config.h>

#include <assert.h>
#include <ctype.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <stddef.h>

#include <getopt.h>
#include <gsl/gsl_rng.h>
#include <gsl/gsl_randist.h>
#include <gsl/gsl_sf.h>

#include "cbt_util.h"
#include "hmmer-io.h"

/* ********************************************************
   CONSTANTS
******************************************************** */

#define SOURCECONTROL_ID "$Id: second-hmm.c,v 1.28 2003/12/03 20:33:53 gec Exp $"
#define TITLE PACKAGE_STRING
#define TITLE_VERSION (TITLE " (" SOURCECONTROL_ID ")" )

#define COPYRIGHT "Copyright (C) 2003 Gavin E. Crooks <gec@compbio.berkeley.edu>"

#ifndef TRUE
#define TRUE (1)
#endif

#ifndef FALSE
#define FALSE (0)
#endif


#define MAX_LINE_LENGTH (2048)
#define MAX_SEQ_LENGTH (1024+50)

#define AA_TYPES (20) 
#define AA_TYPES_EXT (AA_TYPES+1) /* Standard 20 + X  */

// Use alphabetic aa codes for consistency with HMMER
//const char aaCodes[] = "ARNDCQEGHILKMFPSTWYVX";
const char aaCodes[] = "ACDEFGHIKLMNPQRSTVWYX";  /* STAR */

// #define BITS_PER_NAT (1.4426950)

// Basic secondary structure types
// E: strand ("Extended")
// H: Helix
// L: Other  (everything else)
// X: unassigned
// Lower letter: less confidence
#define SS_TYPES (3)
#define SS_TYPES_EXT (SS_TYPES+1)
#define SS_TYPES_MIXED (SS_TYPES_EXT*2)
const char ssCodes[] = "EHLXehlx";

// Translation of STRIDE or DSSP 
// alphabets into three letter alphabet
//
// alphabet:  full dssp+stride alphabets
// translate_fa:  Frishman & Argos, Proteins, 1997
// Translate_ehl: This is the translation used by EVA
//                and casp. It seems to becoming standard.
// 
// E (strand), H (helix), L (other)

const char * ss_alphabet      = "HGIEBbTSC _-LX?";
const char * ss_translate_fa  = "HLLELLLLLLLLLXX";
const char * ss_translate_ehl = "HHHEEELLLLLLLXX";
const char * ss_translate = NULL;


// Maximum secondary structure fragment length and
// fragment types.
// Note: changing fragment length from 7 to 9 results 
// in a 2 fold slowing down simple due to the increase
// in size of various arrays!
#define FRAG_LEN   (9)
#define FRAG_TYPES (19683)  /* 3^9 */
//#define FRAG_LEN   (7)
//#define FRAG_TYPES (2187)  /* 3^9 */
#define STATES     (FRAG_TYPES/3) 

#define CORR_LEN (17)

// Default frag length used in prediction
#define FRAG_LEN_DEFAULT (7)

// Sequences are padded with sentinals ('X') which 
// avoids having to deal with the ends explicetly in
// the prediction algorithm.
#define SEQ_PADDING_LEN ( (int)floor(FRAG_LEN/2))


// Misc. Parameters
//#define ALPHA_DEFAULT (2.5) /*FIXME: REMOVE */
#define BETA_DEFAULT (2.5) 
#define AA_PSEUDOCOUNT (20.) 
#define MIN_SEG_PROB (1.E-6)  /*For print profile */
#define REPLICAS (50)  
#define MINIMUM_CONFIDENCE (0.66)


// Verbosity
#define QUITE   (-1)
#define NORMAL  (0)
#define VERBOSE (1)


/* ********************************************************
   DECLARATIONS
******************************************************** */

// frag: A secondary structure fragment encoded
// as an integer using base 3.
typedef int frag;

// An amino acid encoded as an integer from the 
// range [0, AA_TYPES_EXT)
typedef int amino_acid;

// Various useful conversion routines.
// A 'seq' is an integer array containing
// secondary structure elements encoded as integers
// from the range [0, SS_TYPES) 

int string_to_frag(char * ss_string);
char * frag_to_string(const size_t K, const int f);
int * frag_to_seq(const size_t K, const frag f);
frag seq_to_frag(const size_t  K, const int *seq);

amino_acid char_to_aa(char one_letter_code);
char aa_to_char(amino_acid aa);
int char_to_ss(char one_letter_code);
char ss_to_char(int ss);

char digit_to_char(int digit);
int prob_to_digit(double p);


// Attach dummy residues to each end of profile to 
// match padding of sequences
void pad_profile( aa_profile * aa_prof, int padding) {
  int len = aa_prof->len;
  int i,a;

  for(i = len + padding-1; i >= padding; i--) {
    for(a=0;a <20; a++) {
      aa_prof->profile[i][a] = aa_prof->profile[i-padding][a];
    }
  }

  for(a=0;a <20; a++) {
    for(i = 0; i< padding; i++)
      aa_prof->profile[i][a] = 0.0;
    for(i = len+padding; i<len+2*padding; i++)
      aa_prof->profile[i][a] = 0.0;
  }
  aa_prof->len = len + 2* padding;
}

/* *************************************************** * 
                     Sequence
 * *************************************************** */

typedef struct sequence_str {
  char id[MAX_LINE_LENGTH];
  char comment[MAX_LINE_LENGTH];
  int length;
  int letter[MAX_SEQ_LENGTH];
} * Sequence;

Sequence sequence_create();
int sequence_parse(Sequence self, FILE * stream, int (*char_to_int)(char c)    );
void sequence_print(Sequence self, FILE * stream, char (*int_to_char)(int i) );
void sequence_free(Sequence self);

Sequence sequence_create() {
  Sequence self;
  self = calloc(1,sizeof(struct sequence_str));
  if(self==NULL) cbt_die ("Cannot Allocate space for sequence");
  return self;
}

void sequence_free(Sequence self) {
  if(self!=NULL) free(self);
}

/* 
 * Parse a fasta format file into a sequence.
 * 
 * >sequence_id From_first_white_space_to_EOL_is_just_comments
 * thisisaseqthisisaseqthisisaseqthisisaseqthisisaseqthisisaseq
 * thisisaseqthisisaseqthisisaseqthisisaseqthisisaseqthis 
 */
int sequence_parse(Sequence self, FILE * stream, int (*char_to_int)(char c)    ) {
  int c;
  int length;
  int i;

  c = fgetc(stream);
  if( c==EOF ) return EOF;
  if( c!='>' ) cbt_die("Expected '>', found %c",c);

  c = fgetc(stream);
  if(c != ' ' ) ungetc(c, stream);

  // Read id
  length =0;
  while( !isspace(c=fgetc(stream))) {
    if (c==EOF) cbt_die("Premature EOF");
    self->id[length] =c;
    length +=1;
    if(length>MAX_LINE_LENGTH) cbt_die("Sequence ID is too long!");
  }
  self->id[length] = '\0';

  // read comment
  if(c!='\n') {
    length =0;
    while( (c=fgetc(stream)) !='\n') {
      if (c==EOF) cbt_die("Premature EOF");
      self->comment[length] =c;
      length +=1;
      if(length>MAX_LINE_LENGTH) cbt_die("Sequence comment is too long!");
    }
    self->comment[length] = '\0';
  } else {
    // No comment
    self->comment[0] = '\0';
  }

  // Pad sequence
  for(i=0; i<SEQ_PADDING_LEN; i++) {
    self->letter[i] = (*char_to_int)( 'X');
  }

  length= SEQ_PADDING_LEN;
  while( (c=fgetc(stream))!='>') {
    if (c=='\n' || c==' ') continue;
    if (c==EOF) break;
    self->letter[length] = (*char_to_int)((char)c);
    length +=1;
    if(length>(MAX_SEQ_LENGTH-2*SEQ_PADDING_LEN) )
       cbt_die("Sequence is too long! %d %s", length, self->id );
  }

  // Pad sequence
  for(i=0; i<SEQ_PADDING_LEN; i++) {
    self->letter[length+i] = (*char_to_int)( 'X');
  }
  length += SEQ_PADDING_LEN;




  self->length =length;

  // If this is not the last sequence, then we started to read
  // the id line of the next sequence. Put it back!
  if(c=='>') ungetc('>',stream);

  return length;
}


void sequence_print(Sequence self, FILE * stream, char (*int_to_char)(int i) ) {
  const int line_length = 60;
  int length =self->length;
  int c,l;
  fprintf(stream, ">%s %s\n",self->id, self->comment);

  for(l=0,c=SEQ_PADDING_LEN; c<length-SEQ_PADDING_LEN; c++,l++) {
    if(l==line_length) {
      fprintf(stream, "\n");
      l=0;
    }
    fprintf(stream, "%c", (*int_to_char)(self->letter[c]));
  }
  fprintf(stream, "\n");
}


/* ************************* */




/* *************************************************** * 
                     Profiles
 * *************************************************** */

									 
// P(r|s)
// P(r1, r2 | s1, s2 ; k) / [ P(r1 | s1) P(r2 | s2) ]
// P(s) ... P(s1s2s3s4...)

typedef struct profile_str {
  double prob_aa[AA_TYPES_EXT];
  double prob_ss[SS_TYPES];
  double prob_aa_given_ss[AA_TYPES_EXT][SS_TYPES]; 
  
  double prob_aa2[CORR_LEN][AA_TYPES_EXT][AA_TYPES_EXT];
  double prob_ss2[CORR_LEN][SS_TYPES][SS_TYPES];
  double prob_aa2_ss2[CORR_LEN][AA_TYPES_EXT][AA_TYPES_EXT][SS_TYPES][SS_TYPES];

  double prob_seg[FRAG_LEN+1][FRAG_TYPES];
  double prob_seg_aa[FRAG_LEN+1][FRAG_TYPES][AA_TYPES_EXT];
 
  double total_seg[FRAG_LEN+1];
} * Profile;


Profile profile_create() {
  Profile self = (Profile) calloc(1,sizeof(struct profile_str));
  if(self==NULL) cbt_die("Cannot Allocate");
  return self;
}

void profile_free(Profile self) {
  if(self!=NULL) free(self);
}

int total_count =0;
int count_aa[AA_TYPES];
int count_ss[SS_TYPES];
int count_aa_ss[AA_TYPES][SS_TYPES];
int count_aa2_ss2[CORR_LEN][AA_TYPES][AA_TYPES][SS_TYPES][SS_TYPES];
int total_aa2_ss2[CORR_LEN];
int count_seg[FRAG_LEN+1][FRAG_TYPES];
int total_seg[FRAG_LEN+1];

int count_seg_aa[FRAG_LEN+1][FRAG_TYPES][AA_TYPES_EXT];



void 
profile_analize(Profile self, 
                FILE * fin,
                FILE * fhmm,
                int partition,
                int resample) {
  Sequence aa_seq , ss_seq;
  aa_profile * aa_prof;
  const int SS_X = char_to_ss('X');
  const int AA_X = char_to_aa('X');
  const int CORR_TYPES 
    = AA_TYPES_EXT*AA_TYPES_EXT*SS_TYPES*SS_TYPES;
  
  double w, mean_beta, beta[AA_TYPES];

  int count_ss2[CORR_LEN][SS_TYPES][SS_TYPES];
  int count_aa2[CORR_LEN][AA_TYPES][AA_TYPES];

  int flag;
  int a1,a2,s1,s2;
  int d, i, len, seg;
  int types;

  int seq_count =0;

  // Resampling
  int weight = 1;
  const gsl_rng_type * rng_type;
  gsl_rng * rng = NULL;

  aa_prof = aa_profile_create();

  if(resample) {
    /* Initilize random number generator */
    gsl_rng_env_setup();
    rng_type = gsl_rng_default;
    rng = gsl_rng_alloc(rng_type);
    gsl_rng_set(rng, cbt_ran_clock());
  }


  total_count =0;
  bzero( count_aa, sizeof(int)*AA_TYPES);
  bzero( count_ss, sizeof(int)*SS_TYPES);
  bzero( count_aa_ss,sizeof(int)* AA_TYPES*SS_TYPES);
  bzero( count_aa2_ss2,sizeof(int)* CORR_LEN*AA_TYPES*AA_TYPES*SS_TYPES*SS_TYPES);
  bzero( total_aa2_ss2,sizeof(int)* CORR_LEN );
  bzero( count_ss2,sizeof(int)* CORR_LEN*SS_TYPES*SS_TYPES);
  bzero( count_aa2,sizeof(int)* CORR_LEN*AA_TYPES*AA_TYPES);
  bzero( count_seg, sizeof(int)*(FRAG_LEN+1)*FRAG_TYPES );
  bzero( total_seg ,sizeof(int)*(FRAG_LEN+1) );
  bzero( count_seg_aa ,sizeof(int)*(FRAG_LEN+1)
         *FRAG_TYPES*AA_TYPES_EXT );

  bzero ( self->prob_aa2_ss2, 
          sizeof(double) * CORR_LEN * CORR_TYPES);
  
  aa_seq = sequence_create();
  ss_seq = sequence_create();
  
  while(1) {
    flag = sequence_parse(aa_seq, fin, char_to_aa);
    if(flag==EOF) break;

    flag = sequence_parse(ss_seq, fin, char_to_ss);
    if(flag==EOF) cbt_die("Premature end of file");

    // Load optional homology profile
    if(fhmm != NULL) {
      flag = aa_profile_read_hmmer(aa_prof, fhmm);
      if(flag== EOF) cbt_die("Premature end of hmm library");
      pad_profile(aa_prof, SEQ_PADDING_LEN);
    }


    // Optionally partition data into test and training sets.
    // Training set is every other sequence, starting with first
    seq_count +=1;
    if(partition && ( seq_count%2 ==0) ) continue;

    if(resample) {
      // Bootstrap resampling. Reweight each sequence with
      // a random number drawn from a unit mean Poisson.
      weight = gsl_ran_poisson(rng, 1.0);  
    }

    assert(aa_seq->length == ss_seq->length);
    
    len = ss_seq->length;
    
    for(i=0; i < len; i++) {
      a1 = aa_seq->letter[i];
      s1 = ss_seq->letter[i];
      if(a1==AA_X || s1==SS_X) continue;
      count_aa[a1] += weight;
      count_ss[s1] += weight;
      count_aa_ss[a1][s1] += weight;
      total_count += weight;
    }
    for(d=1; d<CORR_LEN; d++) {
      for(i=0; i < len-d; i++) {
	a1 = aa_seq->letter[i];
	a2 = aa_seq->letter[i+d];
	s1 = ss_seq->letter[i];
	s2 = ss_seq->letter[i+d];
	if(a1 == AA_X || a2 == AA_X || s1==SS_X || s2==SS_X ) continue;
	
	count_aa2_ss2[d][a1][a2][s1][s2] += weight;
	count_ss2[d][s1][s2] += weight;
        count_aa2[d][a1][a2] += weight;
	total_aa2_ss2[d] += weight;
      }
    }

    for(d=1; d<=FRAG_LEN; d++) {
      for(i=0; i <len-d; i++) {
	seg = seq_to_frag(d, ss_seq->letter + i);
	if(seg == -1) continue;
	count_seg[d][seg] += weight;
        total_seg[d] += weight;

        a1 = aa_seq->letter[i + (d/2)];
        count_seg_aa[d][seg][a1] += weight;

      }
    }

		
    
  } // End reading sequences
  
  // Counting done.
  
  for(a1=0; a1<AA_TYPES; a1++)  
    self->prob_aa[a1] = (double) count_aa[a1]/total_count;
  
  for(s1=0; s1<SS_TYPES; s1++)  
    self->prob_ss[s1] = (double) count_ss[s1]/total_count;
  
  for(a1=0; a1<AA_TYPES; a1++) 
    for(s1=0; s1<SS_TYPES; s1++)
      self->prob_aa_given_ss[a1][s1] = 
	( (double) count_aa_ss[a1][s1]/total_count) 
        / self->prob_ss[s1];

  // ss2, aa2
  for(d=0;d<CORR_LEN; d++) {
    for(s1=0; s1<SS_TYPES; s1++)
      for(s2=0; s2<SS_TYPES; s2++) 
        self->prob_ss2[d][s1][s2] = (double) count_ss2[d][s1][s2];
    cbt_normalize(SS_TYPES*SS_TYPES, (double *) self->prob_ss2[d]); 
    
    bzero(self->prob_aa2[d], AA_TYPES_EXT*AA_TYPES_EXT); 
    for(a1=0; a1<AA_TYPES;a1++)
      for(a2=0; a2<AA_TYPES; a2++) 
        self->prob_aa2[d][a1][a2] = (double) count_aa2[d][a1][a2];
    cbt_normalize(AA_TYPES_EXT*AA_TYPES_EXT, (double *) self->prob_aa2[d]); 
  }

  /* Calculate prob_aa2_ss2 */
  for(d=0;d<CORR_LEN; d++){
    for(a1=0; a1<AA_TYPES; a1++) 
      for(a2=0; a2<AA_TYPES; a2++) 
	for(s1=0; s1<SS_TYPES; s1++)
          for(s2=0; s2<SS_TYPES; s2++) 
	    self->prob_aa2_ss2[d][a1][a2][s1][s2] = 
	      (double) count_aa2_ss2[d][a1][a2][s1][s2];
    cbt_normalize(CORR_TYPES, (double *) self->prob_aa2_ss2[d]);  
  }

  // Segment probabilities
  for(d=0;d<=FRAG_LEN; d++) {
    types = (int) pow(SS_TYPES, d);
    for(seg=0; seg<types; seg++) {
      self->prob_seg[d][seg] = (double) count_seg[d][seg];
    }
    cbt_normalize(types, self->prob_seg[d]);
  }


  for(d=0;d<=FRAG_LEN; d++) 
    self->total_seg[d] = total_seg[d];


  for(d=0; d<=FRAG_LEN; d++) {    
    types = (int) pow(SS_TYPES, d);
    for(seg=0; seg<types; seg++) {
      for(a1=0; a1< AA_TYPES; a1++) {
        s1 = frag_to_seq(d, seg)[d/2];
        self->prob_seg_aa[d][seg][a1] = 
          (double) count_seg_aa[d][seg][a1] 
          + AA_PSEUDOCOUNT *
          self->prob_aa_given_ss[a1][s1];
      }
    
      cbt_normalize(AA_TYPES, self->prob_seg_aa[d][seg]);
      self->prob_seg_aa[d][seg][AA_X] = 1.0;
    }
  }

  // Put in 'X'
  for(a1=0; a1<AA_TYPES; a1++)  
    self->prob_aa[AA_X] += self->prob_aa[a1];

  for(a1=0; a1<AA_TYPES; a1++) 
    for(s1=0; s1<SS_TYPES; s1++)
      self->prob_aa_given_ss[AA_X][s1] +=
        self->prob_aa_given_ss[a1][s1];

  for(d=0;d<CORR_LEN; d++) 
    for(a1=0; a1<AA_TYPES; a1++) 
      for(a2=0; a2<AA_TYPES; a2++) 
	for(s1=0; s1<SS_TYPES; s1++)
          for(s2=0; s2<SS_TYPES; s2++) {
	    self->prob_aa2_ss2[d][AA_X][a2][s1][s2] +=
              self->prob_aa2_ss2[d][a1][a2][s1][s2];
	    self->prob_aa2_ss2[d][a1][AA_X][s1][s2] +=
              self->prob_aa2_ss2[d][a1][a2][s1][s2];
	    self->prob_aa2_ss2[d][AA_X][AA_X][s1][s2] +=
              self->prob_aa2_ss2[d][a1][a2][s1][s2];
          }

  return;
}

double
profile_prob_aa_given_seg(const Profile self,
                          const int seg_len,
                          const int aa,
                          const int seg_index) {
  assert( seg_len <= FRAG_LEN && seg_len>=1);

  return
    self->prob_seg_aa[seg_len][seg_index][aa];
}

double
profile_prob_aa_given_seg_prof(const Profile self,
                               const int seg_len,
                               const int aa, /*NOT USED*/
                               const int seg_index,
                               const double * aa_prof,
                               const double beta ) {
  int r1;
  double p, q;
  double relEnt =0.0;
  assert( seg_len <= FRAG_LEN && seg_len>=1);
  
  relEnt = 0.0;
  
  for(r1=0;r1<AA_TYPES; r1++) {
    p = aa_prof[r1];
    q=self->prob_seg_aa[seg_len][seg_index][r1];

    if( p!=0.0 && q != 0.0) relEnt += p * log( p/q);
  }
  
   return exp(-beta*relEnt); // Approximatly normalized
}



// Bootstrap bias to find bias error
void 
bootstrap(const gsl_rng * rng,
          const double obs, 
          const size_t K,
          const double replicas[K],
          double * obs_unbiased,
          double * error,
          double * sd,
          double * bias,
          double * bias_sd) {
  int k, i;
  double total, m, weight;

  double data;
  double r[REPLICAS];

  m = cbt_mean( K, replicas);
  * sd   = cbt_stddev( K, replicas);
  * bias = m - obs ;
  *obs_unbiased = obs - (* bias);
  // Bootstrap bias
  for( i=0; i<REPLICAS; i++) {
    total = 0.0;
    data = 0.0;
    for(k=0; k<K; k++) {
      weight = gsl_ran_poisson(rng, 1.0);        
      total += weight;
      data += weight * replicas[k];
    }
    r[i] = data / total;
  }

  printf(">>> BIAS MEAN %g\n", cbt_mean( REPLICAS, r ));
  *bias_sd = cbt_stddev(REPLICAS, r);
  *error = (*sd) + (*bias_sd);
}



// Calculate various entropies 
// Resamples the profile to do bias correction and
// error estimation.
void profile_stats(FILE * fdata, FILE *fout) {
  int d,r;
  int s,a,a1,a2,s1,s2;
  int resample = FALSE;
  double ent;
  double prob,q;
  double bias, sd;
  int types;

  int t;

  double entropy_ss[REPLICAS];
  double entropy_aa[REPLICAS];
  double entropy_mutual[REPLICAS];
  double seg_entropy[FRAG_LEN+1][REPLICAS];

  double excess_info[CORR_LEN+1][REPLICAS];
  double excess_info_a[CORR_LEN+1][REPLICAS];

  double mut_info_aa[CORR_LEN+1][REPLICAS];  
  double mut_info_aa_cond_ss[CORR_LEN+1][REPLICAS];  

  double seg_ent_mutual[FRAG_LEN+1][REPLICAS];

  Profile prof = profile_create();
  
  for(r=0;r<REPLICAS; r++) {
    rewind(fdata);
    profile_analize(prof, fdata, NULL, FALSE, resample);
    resample = TRUE; // resample all except for first replica

    // Secondary segment entropy
    for(d=1;d<=FRAG_LEN; d++) {
      types = (int) pow(SS_TYPES, d);
      seg_entropy[d][r] = cbt_entropy( types, prof->prob_seg[d]);
    }

    // SS frag to central AA  mutual info. 
    for(d=1;d<=FRAG_LEN; d++) {
      ent =0.0;
      types = (int) pow(SS_TYPES, d);
      for(t=0;t <types; t++) {      
        for(a1=0;a1<AA_TYPES;a1++) {
          prob = prof->prob_seg[d][t] *
            prof->prob_seg_aa[d][t][a1];
          q = prof->prob_seg[d][t] * prof->prob_aa[a1];
          if( prob>0.00000000001)
            ent +=  cbt_bits_per_nat * prob* log(prob/q);
        }
        seg_ent_mutual[d][r] = ent;
      }
    }


    entropy_aa[r] = cbt_entropy( AA_TYPES, prof->prob_aa);
    entropy_ss[r] = cbt_entropy( SS_TYPES, prof->prob_ss);

    // SS  to AA  mutual info. 
    ent= 0.0;
    for(a =0;a<AA_TYPES; a++) {
      for(s =0;s<SS_TYPES; s++) {
        prob = prof->prob_aa_given_ss[a][s] * prof->prob_ss[s];
        q= prof->prob_aa[a] * prof->prob_ss[s];
        ent +=  cbt_bits_per_nat * prob * log(prob/q);
      }
    }
    entropy_mutual[r] = ent;
  
    // Excess mutual info
    for(d=1;d <CORR_LEN; d++) {
      excess_info[d][r] =0.0;
      for(s1 =0;s1<SS_TYPES; s1++) {
        for(s2 =0;s2<SS_TYPES; s2++) {
          for(a1 =0;a1<AA_TYPES; a1++) {
            for(a2 =0;a2<AA_TYPES; a2++) { 
              prob = prof->prob_aa2_ss2[d][a1][a2][s1][s2];
              q = prof->prob_ss2[d][s1][s2] *
                prof->prob_aa2[d][a1][a2];
              if( prob>0.00000000001)
                excess_info[d][r] += cbt_bits_per_nat
                  * prob* log(prob/q);
            }
          }
        }
      }
      excess_info[d][r] += -2.*entropy_mutual[r];
    }

    // Excess mutual information ignoring amino acid correlations
    for(d=1;d <CORR_LEN; d++) {
      excess_info_a[d][r] =0.0;
      for(s1 =0;s1<SS_TYPES; s1++) {
        for(s2 =0;s2<SS_TYPES; s2++) {
          for(a1 =0;a1<AA_TYPES; a1++) {
            q = prof->prob_aa[a1] * prof->prob_ss2[d][s1][s2];

            // left
            prob = 0.0;
            for(a2 =0;a2<AA_TYPES; a2++) 
              prob += prof->prob_aa2_ss2[d][a1][a2][s1][s2];
            if( prob>0.00000000001)
              excess_info_a[d][r] +=  cbt_bits_per_nat 
                * prob* log(prob/q);

            //right
            prob = 0.0;
            for(a2 =0;a2<AA_TYPES; a2++) 
              prob += prof->prob_aa2_ss2[d][a2][a1][s1][s2];
            if( prob>0.00000000001)
              excess_info_a[d][r] +=  cbt_bits_per_nat
                * prob* log(prob/q);
          }
        }
      }
      excess_info_a[d][r] += -2.*entropy_mutual[r];
    }


    // AA to AA mutual info
    for(d=1;d <CORR_LEN; d++) {
      mut_info_aa[d][r] =0.0;
      for(a1 =0;a1<AA_TYPES; a1++) {      
        for(a2 =0;a2<AA_TYPES; a2++) { 
              prob = prof->prob_aa2[d][a1][a2];
              q = prof->prob_aa[a1] *
                prof->prob_aa[a2];
              if( prob>0.00000000001)
                mut_info_aa[d][r] +=  cbt_bits_per_nat
                  * prob* log(prob/q);
        }
      }
    }

    // Conditional mutual information
    for(d=1;d <CORR_LEN; d++) {
      mut_info_aa_cond_ss[d][r] =
        mut_info_aa[d][r] + 
        excess_info[d][r] -
        excess_info_a[d][r];
    }

  } //end replicas


  fprintf(fout,"\n");
  fprintf(fout, "# --- Profile statistics ---\n\n\n");

  ent = entropy_ss[0];
  sd = cbt_stddev( REPLICAS-1, entropy_ss+1);
  bias = cbt_mean( REPLICAS-1, entropy_ss+1)-ent;
  fprintf(fout, "# SS Entropy, S(s) [SD, bias correction] \n");
  fprintf(fout, "%g\t%g\t%g\n\n\n", ent-bias,sd, bias);

  ent = entropy_aa[0];
  sd = cbt_stddev( REPLICAS-1, entropy_aa+1);
  bias = cbt_mean( REPLICAS-1, entropy_aa+1)-ent;
  fprintf(fout, "# Amino Acid Entropy, S(r) [SD, bias correction]\n");
  fprintf(fout, "%g\t%g\t%g\n\n\n", ent-bias, sd, bias);

  ent = entropy_mutual[0];
  sd = cbt_stddev( REPLICAS-1, entropy_mutual+1);
  bias = cbt_mean( REPLICAS-1, entropy_mutual+1)-ent;
  fprintf(fout, "# Mutual Information, I(r;s)  [sd, bias corr.]\n");
  fprintf(fout, "%g\t%g\t%g\n\n\n", ent-bias, sd, bias);

  fprintf(fout, "# Secondary Segment Entropy:\n");
  fprintf(fout, "#  L \tS_L \tS_L/L\t [sd, bias corr.]\n");
  for(d=1; d<=FRAG_LEN; d++) {
    ent = seg_entropy[d][0];
    sd = cbt_stddev( REPLICAS-1, seg_entropy[d]+1);
    bias = cbt_mean( REPLICAS-1, seg_entropy[d]+1)-ent;
    fprintf(fout, "%d\t%g\t%g\t%g\t%g\n",
            d, ent-bias, ent/d, sd, bias);
  }

  fprintf(fout,"\n\n");

  fprintf(fout, "# Excess Mutual Information: I(R1, R2; S1, S2 | k) - I(R1; S1) - I(R2; S2) [sd, bias corr.]\n");
  for(d=1; d<CORR_LEN; d++) {
    ent = excess_info[d][0]; 
    sd = cbt_stddev( REPLICAS-1, excess_info[d]+1);
    bias = cbt_mean( REPLICAS-1, excess_info[d]+1) - ent;
    
    fprintf(fout, "%d\t%g\t%g\t%g\n",
            d, ent-bias, sd, bias );
  }
  fprintf(fout,"\n\n");

  fprintf(fout, "# Excess Mutual Information Ignoring Amino Acid Correlations\n");
  fprintf(fout, "#  I( R1; S1, S2 | k) + I(R2; S1, s2 | k) - I(R1; S1) - I(R2; S2) [sd, bias corr.]\n");
 
  for(d=1; d<CORR_LEN; d++) {
    ent = excess_info_a[d][0]; 
    sd = cbt_stddev( REPLICAS-1, excess_info_a[d]+1);
    bias = cbt_mean( REPLICAS-1, excess_info_a[d]+1) - ent;
    
    fprintf(fout, "%d\t%g\t%g\t%g\n",
            d, ent-bias, sd, bias );
  }
  fprintf(fout,"\n\n");


  fprintf(fout, "# Conditional AA Mutual Information: I(R1; R2| S1, S2, k) [sd, bias corr.]\n");
  for(d=1; d<CORR_LEN; d++) {
    ent = mut_info_aa_cond_ss[d][0]; 
    sd = cbt_stddev( REPLICAS-1, mut_info_aa_cond_ss[d]+1);
    bias = cbt_mean( REPLICAS-1, mut_info_aa_cond_ss[d]+1) - ent;
    
    fprintf(fout, "%d\t%g\t%g\t%g\n",
            d, ent-bias, sd, bias );
  }
  fprintf(fout,"\n\n");



  fprintf(fout, "# Secondary Segment to central amino acid mutual info:\n");
  fprintf(fout, "#  L\tS_L\tsd.\tbias corr. \n");
  for(d=1; d<=FRAG_LEN; d++) {
    ent = seg_ent_mutual[d][0];
    sd = cbt_stddev( REPLICAS-1, seg_ent_mutual[d]+1);
    bias = cbt_mean( REPLICAS-1, seg_ent_mutual[d]+1)-ent;    
    fprintf(fout, "%d\t%g\t%g\t%g\n",
            d, ent-bias, sd, bias);
  }


  fprintf(fout,"\n\n");

  profile_free(prof);
  
  rewind(fdata);
}

void profile_print( Profile self, FILE * fout, int verbosity) {
  int a1,s1;
  //int a2,s2;
  int d,seg;
  int types;

  fprintf(fout, "# --- Secondary Structure Probabilities ---\n\n");

  fprintf(fout, "# P(r)\n\n");
  
  for(a1=0; a1<AA_TYPES; a1++)
    fprintf(fout, "%c %g\n",
	    aa_to_char(a1),
	    self->prob_aa[a1]);
  
  fprintf(fout, "\n\n\n# P(s)\n\n");	
  for(s1=0; s1<SS_TYPES; s1++)
    fprintf(fout, "%c %6g\n",
	    ss_to_char(s1),
	    self->prob_ss[s1]);
  
  fprintf(fout, "\n\n\n#  P(r|s)\n\n");	
  for(a1=0; a1<AA_TYPES; a1++) {
    for(s1=0; s1<SS_TYPES; s1++) {
      fprintf(fout, "%c %c %6g\n",
	      aa_to_char(a1),
	      ss_to_char(s1),
	      self->prob_aa_given_ss[a1][s1]);
    }
  }

  fprintf(fout, "\n# Secondary Segment Probabilities\n");
  for(d=1; d<=FRAG_LEN; d++) {
    fprintf(fout, "\n# Length = %d\n", d);
    types = (int) pow(SS_TYPES, d);
    for(seg = 0; seg< types; seg++) {
      if(self->prob_seg[d][seg]> MIN_SEG_PROB ) {
	fprintf(fout, "%-10s  %-8.5g  ",
		frag_to_string(d, seg),
		self->prob_seg[d][seg]);
        for(a1=0;a1<AA_TYPES; a1++) {
          fprintf(fout, "  %-8.5g",
                  self->prob_seg_aa[d][seg][a1]); 
        }
        fprintf(fout,"\n");

        fprintf(fout, "%-10s  %-8.5g  ",
		frag_to_string(d, seg),
		self->prob_seg[d][seg]);

        fprintf(fout,"\n");
      }
    }
    fprintf(fout, "\n");
  }

  return;
}



/* *************************************************** * 
                        Markov
 * *************************************************** */

typedef struct markov_str {
  Profile profile;

  Sequence aa_seq;
  Sequence ss_seq;

  // Predictions
  Sequence ss_local;
  Sequence ss_global;

  double prob_ss[MAX_SEQ_LENGTH][SS_TYPES];

  // Accumulated Statistics
  int total_seq;
  int total_ss;
  int obs_pred[SS_TYPES_EXT][SS_TYPES_EXT];
  int pred_correct[10];
  int pred_total[10];
  double total_q3;

  double ss_log_likelihood;
  double frag_log_likelihood;
  double total_frag;

  aa_profile * aa_prof;

  // Used for bootstrap resampling
  gsl_rng * rng;

} * Markov;

void markov_clear_stats(Markov self) {
  self->total_seq =0;
  self->total_ss =0;
  self->total_q3 = 0.0;
  bzero(self->obs_pred, sizeof(int)*SS_TYPES_EXT*SS_TYPES_EXT);
  bzero(self->pred_total, sizeof(int)*10);
  bzero(self->pred_correct, sizeof(int)*10);
  self->ss_log_likelihood = 0.0;
  self->frag_log_likelihood = 0.0;
  self->total_frag = 0.0;
}

Markov markov_create(Profile prof) {
  Markov self;
  const gsl_rng_type * rng_type;

  self = calloc(1,sizeof(struct markov_str));
  if(self==NULL) cbt_die("Cannot Allocate space for Markov");

  self->profile = prof;
  self->aa_seq = sequence_create();
  self->ss_seq = sequence_create();
  self->ss_global = sequence_create();
  strcpy(self->ss_global->id, " Global prediction"); 
  self->ss_local = sequence_create();
  strcpy(self->ss_local->id, " Local prediction"); 

  markov_clear_stats(self);

  /* Initialize random number generator */
  gsl_rng_env_setup();
  rng_type = gsl_rng_default;
  self->rng = gsl_rng_alloc(rng_type);
  gsl_rng_set(self->rng, cbt_ran_clock());

  self->aa_prof = aa_profile_create();

  return self;
}


void markov_free(Markov self) {
  assert(self!=NULL);
  free(self->aa_seq);
  free(self->ss_seq);
  free(self->ss_global);
  gsl_rng_free(self->rng);
  aa_profile_free(self->aa_prof);
  free(self);
}


// Private scratch space
double trans_prob[MAX_SEQ_LENGTH][FRAG_TYPES];
double seg_prob[MAX_SEQ_LENGTH][STATES];
double forward[MAX_SEQ_LENGTH][STATES];
double back[MAX_SEQ_LENGTH][STATES];
double prob[MAX_SEQ_LENGTH][STATES];
double prob_ss[MAX_SEQ_LENGTH][SS_TYPES];





void 
markov_predict(Markov self, 
               const int frag_len,
               const double min_confidence,
               const double beta,
               const int stats,
               const int resample,
               const int use_aa_profile) {
  int a;
  int index, m, aa;
  int next, prev, trans, t ;

  double p;

  int trace;       // global traceback
  double pt[3];

  int ss_obs, ss_pred;
  int weight = 1;

  const int SS_X = char_to_ss('X');
  
  Profile profile= self->profile;

  // Sequence length  
  const int seq_length = self->aa_seq->length; 
  // sequence length of each state
  const int state_len = frag_len-1;      
  // Markov model length
  const int len = seq_length - state_len+1;   
  // Fragment types
  const int frag_types =(int) pow( SS_TYPES, frag_len); 
  // Markov model states
  const int states = frag_types/3;            

  const int half_frag_len = (int) floor( 0.5*frag_len);

  // init
  bzero( forward, sizeof(double)*MAX_SEQ_LENGTH*STATES);
  bzero( back,    sizeof(double)*MAX_SEQ_LENGTH*STATES);
  bzero( seg_prob,sizeof(double)*MAX_SEQ_LENGTH*STATES);
  bzero( trans_prob, sizeof(double)*MAX_SEQ_LENGTH*FRAG_TYPES);
  
  // calculate all state probabilities
  for(index=0;index<states; index++) { 
    p = profile->prob_seg[state_len][index];
    
    for(m=0;m<len;m++) 
      seg_prob[m][index] = p;
  }
  
  for(m=0;m<len;m++) {
    cbt_normalize(states, seg_prob[m]); 
  }

  // Transition probabilities 
  for(index=0;index<frag_types; index++) { 
    p = profile->prob_seg[frag_len][index];
    
    for(m=0;m<len-1;m++) {
      trans_prob[m][index] = p;

      if(p>0.0) {
        // Add primary sequence information
        aa = self->aa_seq->letter[m+half_frag_len];
        if(use_aa_profile == FALSE) {
          trans_prob[m][index] *=         
            profile_prob_aa_given_seg(profile, frag_len, aa, index);
        } else {
          trans_prob[m][index] *= 
            profile_prob_aa_given_seg_prof(profile, 
                                           frag_len, 
                                           aa, 
                                           index,
                                           self->aa_prof->profile[m+half_frag_len], 
                                           beta);

        }
      }
    } 
  }

  for(m=0;m<len-1;m++) {
    cbt_normalize(frag_types, trans_prob[m]);
  }
  

  // Make global prediction by joining the local segements
  // using the Markov Chain and the Forward-Backward algorithm.
  
  // March Forward
  for(index=0;index<states; index++) 
    forward[0][index] = seg_prob[0][index];

  for(a=1; a<len; a++) {
    for(next=0;next<states; next++) {
      forward[a][next] = 0.0;
      
      for(t=0;t<3; t++) {
        trans = next + (t * states);
        prev = trans/3;

        if(seg_prob[a-1][prev] != 0.0 ) {
          forward[a][next] 
            += trans_prob[a-1][trans] 
            * forward[a-1][prev]
            / seg_prob[a-1][prev];
        }
      }
    }
    cbt_normalize(states, forward[a]) ;
  }

  // March Back
  for(index=0;index<states; index++) 
    back[len-1][index] = seg_prob[len-1][index];

  for(a=len-2; a>=0; a--) {
    for(prev=0;prev<states; prev++) {
      back[a][prev] = 0.0;
      
      for(t=0;t<3; t++) {
        trans = (prev*3) + t;
        next = trans % states;

        if(seg_prob[a+1][next] != 0.0 ) {
          back[a][prev] += trans_prob[a][trans] * back[a+1][next] 
            / seg_prob[a+1][next];     
        }
      }
    }
    cbt_normalize(states, back[a]) ;
  }

  // Calculate all Markov model fragment probabilities
  for(a=0; a<len; a++) {
    for(index=0;index<states; index++) { 
      if(seg_prob[a][index] !=0.0) 
	prob[a][index] = forward[a][index] * back[a][index] / seg_prob[a][index];
    }
    cbt_normalize(states, prob[a]);
  }

  // Extract per site secondary structure probabilities
  for(a=0;a<seq_length;a++) {
    for(t=0;t<3;t++) {
      prob_ss[a][t]=0.0;
    }
  }

  for(a=0;a<len;a++) {
    for(index=0; index<states; index++) {
      t = frag_to_seq(state_len, index)[0];
      prob_ss[a][t] += prob[a][index];
    }
    cbt_normalize(3, prob_ss[a]);
  }
  for(a=len; a<seq_length; a++) {
    for(index=0; index<states; index++) {
      t = frag_to_seq(state_len, index)[a-len+1];
      prob_ss[a][t] += prob[len-1][index];
    }
    cbt_normalize(3, prob_ss[a]);
  }

  // Make local secondary structure prediction
  for(a=0; a<seq_length; a++) {
    for(t=0; t<SS_TYPES; t++)
      self->prob_ss[a][t] = prob_ss[a][t];

    ss_pred = cbt_argmax_dbl(3, prob_ss[a]);
    self->ss_local->letter[a]=ss_pred;
    
    // Convert to lower case if not confident of
    // prediction.
    if( prob_ss[a][ss_pred] < min_confidence) { 
      self->ss_local->letter[a] += SS_TYPES_EXT;
    }
  }
  self->ss_local->length = seq_length;
  
  
  // Trace: find optimal global prediction
  trace = cbt_argmax_dbl(states, back[0]);
  self->ss_global->letter[0] =
    frag_to_seq(state_len, trace)[0];

  for(a=1;a<len;a++) {
    pt[0] = back[a][(trace*3 + 0)%states];		
    pt[1] = back[a][(trace*3 + 1)%states];		
    pt[2] = back[a][(trace*3 + 2)%states];
    
    trace= (trace*3 +cbt_argmax_dbl(3, pt) ) %states;
    self->ss_global->letter[a] =
      frag_to_seq(state_len, trace)[0];
  }

  for(a=len;a<seq_length; a++) { 
    self->ss_global->letter[a] =
      frag_to_seq(state_len, trace)[a-len+1];
  }
  
  self->ss_global->length = seq_length;


  // Statistics:
  // If the statistics flag is set then accumulate
  // statistics. self->ss_seq must be correct secondary structure
  if(stats) {
    int total_corr = 0;
    int total_pred = 0;
    double conf = 0.0;
    int s;
    frag frag_obs;

    if(resample) {
      // Bootstrap resampling. Reweight each sequence with
      // a random number drawn from a unit mean Poisson.
      weight = gsl_ran_poisson(self->rng, 1.0);        
    } else {
      weight = 1;
    }

    if(weight !=0) {
      self->total_seq += weight;
      for(a=0; a<seq_length; a++) {

        // Fragment log likelihood
        frag_obs = seq_to_frag(state_len, self->ss_seq->letter+a);
        if(frag_obs !=-1 && prob[a][frag_obs] !=0.0) { // No 'X's
          self->frag_log_likelihood += 
            1.0*weight*cbt_bits_per_nat * log( prob[a][frag_obs]);
          self->total_frag += weight;
        }

        ss_obs = self->ss_seq->letter[a];
        ss_pred = self->ss_local->letter[a];
      
        // Remove confidence encoding
        if(ss_pred>= SS_TYPES_EXT) ss_pred -=SS_TYPES_EXT;
        if(ss_obs == SS_X || ss_pred==SS_X) continue;
  
        self->total_ss += weight;        
        self->obs_pred[ss_pred][ss_obs] += weight;
        
        total_pred += weight;
        if(ss_pred==ss_obs) total_corr +=weight;
        
        conf = prob_ss[a][ss_obs];
        self->pred_correct[ prob_to_digit(conf) ] += weight;
        for(s=0;s<SS_TYPES;s++)
          self->pred_total[ prob_to_digit(prob_ss[a][s]) ] 
            += weight;

        if(prob_ss[a][ss_obs]!=0.0) {
          self->ss_log_likelihood += 
            1.0* weight* cbt_bits_per_nat * log(prob_ss[a][ss_obs]);  
        }

  
      }

      if( total_corr >0) {         
        self->total_q3 += (double)1.0*weight* total_corr/total_pred;
      }
    }
  } //end stats
} // End markov_predict



void 
markov_print_prediction(Markov self, 
                        FILE * fout,
                        const int verbosity,
                        const int stats)
{
  Sequence seq;
  int i, length;

  sequence_print(self->aa_seq, fout, aa_to_char);
  if(stats)
    sequence_print(self->ss_seq, fout, ss_to_char);
  sequence_print(self->ss_local, fout, ss_to_char);

  if(verbosity==VERBOSE) {
    sequence_print(self->ss_global, fout, ss_to_char);

    seq = sequence_create();
    length = self->aa_seq->length;
    seq->length = length;

    strcpy( seq->id, " Helix probability * 10");  
    for(i=0;i<length; i++) 
      seq->letter[i] = prob_to_digit(self->prob_ss[i][0]);  
    sequence_print(seq, fout, digit_to_char);

    strcpy( seq->id, " Strand probability * 10");  
    for(i=0;i<length; i++) 
      seq->letter[i] = prob_to_digit(self->prob_ss[i][1]);  
    sequence_print(seq, fout, digit_to_char);

    strcpy( seq->id, " Coil probability * 10");  
    for(i=0;i<length; i++) 
      seq->letter[i] = prob_to_digit(self->prob_ss[i][2]);  
    sequence_print(seq, fout, digit_to_char);

    sequence_free(seq);
  }
  fprintf(fout, "\n");
}


void 
markov_print_statistics(Markov self, 
                        FILE * fout,
                        const int verbosity) {
  int i,j,d;

  fprintf(fout, "\n");
  fprintf(fout, "# --- Result Statistics ---\n\n");
  
  fprintf(fout, "# Total Sequence: %d\n", self->total_seq);
  fprintf(fout, "# Total Residues: %d\n", self->total_ss);
  fprintf(fout, "# Average Q3 per sequence: %g\n", 
          (double)self->total_q3/self->total_seq);
  fprintf(fout, "# Average Q3 per residue: %g\n\n", 
          ((double)self->obs_pred[0][0]
           + self->obs_pred[1][1] + 
           self->obs_pred[2][2])/self->total_ss );

  fprintf(fout, "# Observed versus predicted secondary structure\n");
  fprintf(fout, "# Obs.  Pred->\n\t");
  for(j=0; j<SS_TYPES; j++) 
      fprintf(fout, "%c\t", ss_to_char(j) );
      
  for(i=0;i<SS_TYPES;i++) {
    fprintf(fout, "\n %c\t", ss_to_char(i));
    for(j=0;j<SS_TYPES;j++) {
      fprintf(fout, "%d\t", self->obs_pred[i][j]);
    }
  }
  fprintf(fout, "\n\n");

  fprintf(fout, " # Prediction Accuracy \n");
  for(d=0;d<10; d++) { // digits
    fprintf(fout, "%d%% - %d%% \t%g\n", d*10, (d+1)*10,
            1.0* self->pred_correct[d] / self->pred_total[d]);
  }
  fprintf(fout, "\n");
  
  fprintf(fout, " # Average Observation Log Likelihood\n"); 
  fprintf(fout, " # Per Residue, Per Segment (frag length-1)\n");
  fprintf(fout, "%g\t%g \n\n", self->ss_log_likelihood/self->total_ss,
          self->frag_log_likelihood/self->total_frag);

  fprintf(fout, "\n");
}




/*********************************************************
 * Misc. Utility routines 
 *********************************************************/

/* 
 * Convert an Amino acid one letter code into an integer
 * between 0 (A, Alanine) and 20 (X, Unknown)
 */
amino_acid char_to_aa(char one_letter_code) {
  int i;
  char c = toupper(one_letter_code);

  // Replace degenerate symbols with concrete examples
  if(c=='B') c='D';
  if(c=='Z') c='E';

  for(i=0;i< AA_TYPES_EXT; i++)
    if(aaCodes[i]== c) return i;
  cbt_die("Unknown Amino Acid Code: %c", one_letter_code);
  return -1;
}

/* 
 * Convert an AminoAcid (an int) derived from aatoi() back into
 * a character.
 */
char aa_to_char(amino_acid aa) {
  assert(aa>=0 && aa<AA_TYPES_EXT);
  return aaCodes[aa];
}

int char_to_ss(char one_letter_code) {
  int i;
  char c = toupper(one_letter_code);

  if(ss_translate==NULL) ss_translate = ss_translate_ehl; // Enforce default

  // Translate into 3 letter code
  for(i=0; ss_alphabet[i] !='\0'; i++) {
    if( ss_alphabet[i]== c) {
      c = ss_translate[i];
      break;
    }
  }

  // Convert to integer
  for(i=0;i< SS_TYPES_EXT; i++)
    if(ssCodes[i]== c) return i;
  cbt_die("Unknown SS Code: %c", one_letter_code);
  return -1;
}

char ss_to_char(int ss) {
  assert(ss>=0 && ss<SS_TYPES_MIXED);
  return ssCodes[ss];
}

int prob_to_digit(double p) {
  assert( p >=0.0 && p <=1.0);
  if( p==1.0) return 9;
  return (int) floor(10.* p);
}


char digit_to_char(int digit) {
  char digits[] = "0123456789";
  assert(digit>=0 && digit<=9);
  return digits[digit];
}



int string_to_frag(char *  ss_string) {
  int k,i=0;
  int X = char_to_ss('X');
  char c = ss_string[0];
  int code=0;

  while( c != '\0') {
    k = char_to_ss(c);
    if( k == X ) cbt_die ("Code for Unkown SS not usable here");
    code *= SS_TYPES ;
    code +=k ;
    i++;
    c = ss_string[i];
  }

  assert( code<FRAG_TYPES );

  return code;
}


char fragStrings[FRAG_LEN+1][FRAG_TYPES][FRAG_LEN+1];
int frag_to_string_init = FALSE;

char * frag_to_string(const size_t K, const int ss) {
  int j,k,s,t, types;
  assert(ss>=0 && ss<FRAG_TYPES);

  // Initilize
  if(!frag_to_string_init) {
    frag_to_string_init = TRUE;
    for(k=0; k<=FRAG_LEN;k++) {
      types = (int) pow(SS_TYPES, k);
      for(t=0; t<types; t++) {
	s= t;
	for(j=k-1; j>=0; j--) {
	  fragStrings[k][t][j] = ss_to_char( s % 3);
          s /= 3;
	}
	fragStrings[k][t][k] = '\0';
        assert( t == string_to_frag(fragStrings[k][t]));  
      }
    }
  }
	
  return fragStrings[K][ss];
}


int frag_seq[FRAG_LEN+1][FRAG_TYPES][FRAG_LEN+1];
int frag_to_seq_init = FALSE;

int * frag_to_seq(const size_t K, const frag f) {
  int j,k,s,t, types;
  assert(f>=0 && f<FRAG_TYPES);

  //Initialize
  if(frag_to_seq_init==FALSE) {
    frag_to_seq_init = TRUE;
    for(k=0; k<=FRAG_LEN;k++) {
      types = (int) pow(SS_TYPES, k);
      assert(types <=FRAG_TYPES);
      for(t=0; t<types; t++) {
	s= t;
	for(j=k-1; j>=0; j--) {
	  frag_seq[k][t][j] =  s % 3;
          s /= 3;
	}
        assert( t == seq_to_frag(k, frag_seq[k][t]) );
      }
    }
  }
	
  return frag_seq[K][f];
}

frag seq_to_frag(const size_t K, const int * seq) {
  int k, f=0;
  const int SS_X = char_to_ss('X');

  for(k=0;k<K;k++) {
    if(seq[k] == SS_X) return -1;
    assert(seq[k] < SS_TYPES);
    f = (f*SS_TYPES) + seq[k];
  }
  return f;
}



void 
predict(FILE * fdata, 
        FILE * fhmm,
        FILE * fin,
        FILE * fout,
        int frag_len,
        double min_confidence,
        const double beta,
        int verbosity,
        int partition,
        int stats, 
        int resample) 
{

  Profile prof;
  Markov markov;
  int seq_count = 0;
  int flag;
  int replicas;
  int resampling = FALSE;
  int r;

  double res, sd, bias;
  double seq_q3[REPLICAS+1];
  double res_q3[REPLICAS+1];
  double ss_ll[REPLICAS+1];
  double frag_ll[REPLICAS+1];


  if(resample) {
    replicas = REPLICAS+1;
  } else {
    replicas = 1;
  }

  prof = profile_create();
  markov = markov_create(prof);

  for(r=0; r<replicas; r++) {
    rewind(fdata);
    if(fhmm!=NULL) rewind(fhmm);
   
    profile_analize(prof, fdata, fhmm, partition, resampling);  

    
    if(r>0) 
      fprintf(fout, "\n# --- Replica %d ---\n", r);

    if(verbosity != QUITE)     
      fprintf(fout, "\n# --- Predictions ---\n\n");

    rewind(fin);
    if(fhmm!=NULL) rewind(fhmm);
    seq_count=0;
    while(sequence_parse(markov->aa_seq, fin, char_to_aa) !=EOF) {
      if(stats) {
        flag = sequence_parse(markov->ss_seq, fin, char_to_ss);
        if(flag==EOF) cbt_die("Missing secondary structure sequence");
      }

      if(fhmm != NULL) {
        if(verbosity != QUITE)     
          fprintf(fout, "HMM ID: %s\n", markov->aa_seq->id);  
        flag = aa_profile_read_hmmer(markov->aa_prof, fhmm);
        pad_profile(markov->aa_prof, SEQ_PADDING_LEN);
        // if(flag== EOF) cbt_die("Premature end of hmm library");
 
        //flag = aa_profile_read_hmmer2(markov->aa_prof, markov->aa_seq->id);

        if(flag==TRUE) {
          printf("  FAILED\n");
          continue;
        }
      }

      // Optionally partition data into test and training sets.
      // Test set is every other sequence, starting with second
      seq_count +=1;
      if(partition && ( seq_count%2 ==1) ) continue;

      // A few multiple alignments from the Barton
      // dataset are inconsistent. Skip them.
      if(fhmm != NULL && 
         (markov->aa_prof->len)!=
         (markov->aa_seq->length)) {
        fprintf(stderr, "Inconsistent HMM. Skipping %s\n",
                markov->aa_seq->id);
        continue;
      }

      // Finally, predict the secondary structure!
      markov_predict(markov, frag_len, min_confidence, 
                     beta, 
                     stats, resampling, fhmm!=NULL);

      if(verbosity != QUITE)     
        markov_print_prediction(markov, fout, verbosity, stats);

      //if(seq_count>50) break; // LIMIT
    } // End predictions


    if(stats) {
      markov_print_statistics(markov, fout, verbosity);

      // Accumulate replica samples
      seq_q3[r] = (double)markov->total_q3/markov->total_seq;
      res_q3[r] =((double) markov->obs_pred[0][0]
                  + markov->obs_pred[1][1] + 
                  markov->obs_pred[2][2])/markov->total_ss ;
      ss_ll[r] = markov->ss_log_likelihood/markov->total_ss;
      frag_ll[r] = markov->frag_log_likelihood/markov->total_frag;
     
      markov_clear_stats(markov);
    }   
    // Resample everything but zeroth replica
    resampling = TRUE;
  } // End replicas


  // Bootstrapped statistics
  if(stats && resample) {
    fprintf(fout,"\n");
    fprintf(fout, "# --- Resampled Statistics ---\n\n");

    res  = seq_q3[0];
    sd   = cbt_stddev( REPLICAS, seq_q3+1);
    bias = cbt_mean( REPLICAS, seq_q3+1)- res;
    fprintf(fout, "# Average Q3 per sequence: [ SD, bias corr.]\n"); 
    fprintf(fout, "%g\t%g\t%g\n\n", res-bias,sd, bias);

    res  = res_q3[0];
    sd   = cbt_stddev( REPLICAS, res_q3+1);
    bias = cbt_mean( REPLICAS, res_q3+1)- res;
    fprintf(fout, "# Average Q3 per residue: [ SD, bias corr.]\n"); 
    fprintf(fout, "%g\t%g\t%g\n\n", res-bias,sd, bias);

    res  = ss_ll[0];
    sd   = cbt_stddev( REPLICAS, ss_ll+1);
    bias = cbt_mean( REPLICAS, ss_ll+1)- res;
    fprintf(fout, "# Average Observation Log Likelihood per Residue [SD, bias correction.]\n"); 
    fprintf(fout, "%g\t%g\t%g\n\n", res-bias,sd, bias);

    res  =frag_ll[0];
    sd   = cbt_stddev( REPLICAS, frag_ll+1);
    bias = cbt_mean( REPLICAS, frag_ll+1)- res;
    fprintf(fout, "# Average Observation Log Likelihood per Segment [SD, bias correction.]\n"); 
    fprintf(fout, "%g\t%g\t%g\n\n", res-bias, sd, bias);
  }



  markov_free(markov);
  profile_free(prof);
}


/* ********************************************************
   MAIN
******************************************************** */

void help( void ) {
  printf("\n"
         "%s\n%s\n"
         "\n"
         "Usage: second-hmm -d data/data.fa -i data/test.fa\n"
         "\n"
         "Description:\n"
         "  Predict protein secondary structure from an amino acid\n" 
         "  sequence using amino acid profiles of short secondary\n"
         "  structure fragments and a hidden Markov model.\n"
         "\n"
         " Options:\n"
         "\n"
         "     --beta  <float>       Profile scoring parameter (default %g)\n"
         "  -c --confidence <float>  Minimum confidence for secondary structure\n" 
         "                           prediction. Confident predictions are in upper\n"
         "                           case.  (default: %g)\n" 
         "  -d --data    <filename>  Training data file\n"
         "  -f --fragment <int>      Fragment length (default: %d)\n"
         "  -h --help                Print this help message\n" 
         "     --hmm     <filename>  HMMER profiles for input sequences.\n"
         "  -i --input   <filename>  Input file (use '@' for stdin)\n"
         "     --mapping <map name>  Select the mapping of DSSP/STRIDE secondary\n"
         "                           structure alphabets to the three letter alphabet\n"
         "                           E (extended beta sheet), H (helix) and L (other).\n"
         "                   EHL     The CASP/EVA mapping. (Default)\n"
         "                   CK      Chandonia & Karplus mapping\n"
         "                   FA      Frishman & Argos mapping (Same as CK)\n"
         "  -o --output  <filename>  Output file (default: stdout)\n"
         "  -q --quite               Minimal output. Do not output predictions.\n"
         "     --partition           Split datafile into training and test sets\n"
         "                           Implies --stats and preempts --input\n"
         "  -p --profile             Output profile data\n"
         "     --resample            Bootstrap resample the test and training data\n"
         "  -s --stats               Calculate statistics. Requires that the\n" 
         "                           input file contains datafile formated data.\n"
         "  -v --verbose             verbose output\n"
         "  -V --version             print version information\n"
         "\n"
         " Sequences are accepted in fasta format. The datafile should consist of\n"
         " alternating amino acid and secondary structure sequences. The input file\n"
         " should consist of one or more amino acid sequences, unless the stats \n"
         " option is chosen, in which case it should have the same format as the \n"
         " datafile.\n"
          "\n", TITLE_VERSION, COPYRIGHT, 
         BETA_DEFAULT, MINIMUM_CONFIDENCE, FRAG_LEN_DEFAULT
         );
  exit ( 0 );
}

void version(void) {
  printf("%s\n", TITLE_VERSION);
  printf("%s\n", COPYRIGHT);
  exit(0);
}

int main(int argc, char *argv[]) {
  int c, flag;

  // Various parameters that can be set on the command line
  char * infile = NULL;
  char * datafile = NULL;
  char * outfile = NULL;
  char * hmmfile = NULL;

  FILE * fin = NULL;
  FILE * fout = stdout;
  FILE * fdata = NULL;
  FILE * fhmm = NULL;

  int verbosity = NORMAL;
  int frag_len = FRAG_LEN_DEFAULT;
  int display_profile = FALSE;
  int stats = FALSE;
  int partition = FALSE;
  int resample = FALSE;
  double min_confidence = MINIMUM_CONFIDENCE;
  double beta  = BETA_DEFAULT;

  Profile prof;

  int opt_index=0;
  char short_opt[] = "hi:o:qvVd:pf:sc:";
  static struct option long_opt[] = {
    {"help",      no_argument,       NULL, 'h'},
    {"usage",     no_argument,       NULL, 'h'},
    {"input",     required_argument, NULL, 'i'},
    {"output",    required_argument, NULL, 'o'},
    {"quite",     no_argument,       NULL, 'q'},
    {"verbose",   no_argument,       NULL, 'v'},
    {"version",   no_argument,       NULL, 'V'},
    {"data",      required_argument, NULL, 'd'},
    {"partition", no_argument,       NULL, 'P'},
    {"profile",   no_argument,       NULL, 'p'},
    {"fragment",  required_argument, NULL, 'f'},
    {"resample",  no_argument,       NULL, 'r'},
    {"stats",     no_argument,       NULL, 's'},
    {"confidence",required_argument, NULL, 'c'},
    {"mapping",   required_argument, NULL, 'm'},
    {"hmm",       required_argument, NULL, 'H'},
     {"beta",      required_argument, NULL, 'B'},
    {0, 0, 0, 0}
  };

  while(1) {
    c = getopt_long(argc, argv, short_opt, long_opt, &opt_index);
    if(c==-1) break;

    switch (c) {
    
    /* Standard options */
    case 'h' :
      help();
      break;
    case 'i' :
      infile = optarg;
      if(infile[0] == '@') {
        infile = "from stdin";
        fin = stdin;
      } else {
        fin = fopen(optarg, "r");
        if(fin == NULL) cbt_die("Cannot open file %s for input.", optarg);
      }
      break;
    case 'o' :
      outfile = optarg;
      fout = fopen(optarg, "w");
      if(fout == NULL) cbt_die("Cannot open file %s for output.", optarg);
      break;
    case 'q': 
      verbosity = QUITE;
      break;
    case 'v' : 
      verbosity = VERBOSE;
      break;
    case 'V' : 
      version();
      break;
      
    /* Extended options */
    case 'f' :
      flag = sscanf(optarg,"%d",&frag_len);
      if(flag==EOF) cbt_die("Cannot parse fragment length");
      if(frag_len<0 || frag_len > FRAG_LEN )
        cbt_die("Invalid fragment length: %d", frag_len);
      break;
    case 'd' : 
      datafile = optarg;
      break;
    case 'P': // Partition
      partition = TRUE;
      stats = TRUE;
      break;
    case 'p':
      display_profile = TRUE;
      break;
    case 'r': // Resample
      resample = TRUE;
    case 's':
      stats = TRUE;
      break;
    case 'c':
      flag = sscanf(optarg,"%lf",&min_confidence);
      if(flag==EOF) cbt_die("Cannot parse minimum confidence");
      if(min_confidence<0.0 || min_confidence > 1.0)
        cbt_die("Invalid confidence: %f", min_confidence);
      break;
    case 'B':
      flag = sscanf(optarg,"%lf",&beta);
      if(flag==EOF) cbt_die("Cannot parse beta");
      if(beta<0.0)
        cbt_die("Invalid beta: %f", beta);
      break;
    case 'm':
      if( strncmp( optarg, "FA", 2)==0 || strncmp( optarg, "CK", 2)==0)  {
        ss_translate = ss_translate_fa; 
      } else if (strncmp(optarg, "EHL", 3) ==0) {
        ss_translate = ss_translate_ehl;
      } else {
        cbt_die("Invalide mapping name (Valid options are 'CK', 'FA' and 'EHL'): %s", optarg);
      }
      break;
    case 'H' :
      hmmfile = optarg;
      fhmm = fopen(optarg, "r");
      if(fhmm == NULL) cbt_die("Cannot open file %s for input.", optarg);
      break;

    default :
      help();
    }
  }

  if(datafile==NULL) 
    cbt_die("No data file given.", optarg);
  fdata = fopen(datafile, "r");
  if(fdata == NULL) 
    cbt_die("Cannot open data file %s for input.", optarg);

  if(partition) {
    if(fin !=NULL)
      cbt_die("Option --partition preempts --input");
    infile = "Partitioning datafile";
    fin = fdata;
  } 

  assert(fout !=NULL);
  assert(fdata !=NULL);

  // End argument parsing


  // Print header
  fprintf(fout, "# %s\n", TITLE_VERSION);
  fprintf(fout, "# Prediction of protein secondary structure from amino acid sequence\n");
  fprintf(fout, "# Database:   %s\n",datafile);
  if(fin!=NULL)
    fprintf(fout, "# Query file: %s\n", infile);
  if(ss_translate==ss_translate_fa) {
    fprintf(fout, "# Secondary structure translation: Frishman & Argos (CK)\n");
  } else {
    fprintf(fout, "# Secondary structure translation: CASP/EVA (EHL)\n");
  }
  fprintf(fout, "# Fragment length: %d\n", frag_len);
  fprintf(fout, "# Minimum confidence: %g (Upper case predictions)\n",
          min_confidence);
  if(resample || display_profile) 
    fprintf(fout, "# Resampling sequences: %d Replicas \n", REPLICAS);
  if(fhmm!=NULL) {
    fprintf(fout, "# HMM library:   %s\n", hmmfile);
    fprintf(fout, "# beta: %g\n", beta);
  }
  fprintf(fout, "\n");


  // Generate and print profile and profile statistics
  if(display_profile) {
    if(verbosity != QUITE) {
      prof = profile_create();
      profile_analize(prof, fdata, fhmm,  partition, resample);  
      profile_print(prof, fout, verbosity);
      free(prof);
    }
    profile_stats(fdata, fout);
  }

  if(fin == NULL ) {
    fprintf(fout, "# --- END %s (No query file.) ---\n\n", TITLE);
    exit(0);
  }

  // Make predictions
  predict(fdata, fhmm, fin, fout, frag_len, min_confidence, 
          beta,
          verbosity, partition, stats, resample);

  fprintf(fout, "# --- END %s ---\n\n", TITLE);
  return 0;
}


